/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.variability.analysis;

import java.util.List;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.variability.model.features.configuration.VariantConfiguration;

import com.microsoft.z3.BoolExpr;
import com.microsoft.z3.Expr;

/**
 * Interface for product line constraints to be checked by the {@link ProductLineBaseAnalysis}.
 * 
 * Note that usually this interface should not be implemented immediately. Instead use the base
 * class {@link ProductLineConstraintBase} to define new constraints.
 * 
 * @author bayha
 */
public interface IProductLineConstraint {

	/**
	 * The name of the constraint.
	 * 
	 * @return The name as {@link String}.
	 */
	public String getConstraintName();

	/**
	 * Creates a meaningful error message for violations of this constraint.
	 * 
	 * @param violatingObjects
	 *            The {@link EObject}s which violate the constraint in combination.
	 * @param violatingConfiguration
	 *            A {@link VariantConfiguration} which will result in a model variant in which the
	 *            'violatingObjects' will violate this constraint.
	 * @return The error message as a {@link String}.
	 */
	public String createErrorMessage(List<EObject> violatingObjects,
			VariantConfiguration violatingConfiguration);

	/**
	 * Retrieves the quantified variables for this constraint.
	 * 
	 * @return An array of variables as {@link Expr}s.
	 */
	public Expr[] getQuantifierVariables();

	/**
	 * Retrieves the constrain body for this constraint.
	 * 
	 * @return The constraint body as a {@link BoolExpr}.
	 */
	public BoolExpr getBody();

	/**
	 * Specifies whether this constraint is quantified as "for all".
	 * 
	 * @return 'true' if this constraint is quantified as "for all". 'false' if "exists".
	 */
	public boolean isForAll();

	/**
	 * Retrieves the generated tracker variables to identify violating interpretations for the
	 * quantifier variables.
	 * 
	 * @return An array of {@link Expr}s with the quantifier variables.
	 */
	public Expr[] getVariableTracker();

	/**
	 * Retrieves the {@link EClass} types for the quantified variables.
	 * 
	 * @return An array of {@link EClass}s for all quantified variables.
	 */
	public EClass[] getVariablesClasses();
}
