/*--------------------------------------------------------------------------+
| Copyright 2021 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.variability.util;

import static java.util.stream.Collectors.toList;
import static org.fortiss.variability.model.VariabilityModelElementFactory.createFeatureConfigurationForFeature;
import static org.fortiss.variability.util.VariabilityUtilsInternal.getAllReferences;
import static org.fortiss.variability.util.VariabilityUtilsInternal.getParentsWithType;

import java.util.Collection;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

import org.eclipse.emf.common.util.EList;
import org.fortiss.variability.model.features.AbstractCompositionalFeature;
import org.fortiss.variability.model.features.AbstractCrossFeatureConstraint;
import org.fortiss.variability.model.features.AbstractFeature;
import org.fortiss.variability.model.features.configuration.CompositionalFeatureConfiguration;
import org.fortiss.variability.model.features.configuration.IFeatureConfiguration;
import org.fortiss.variability.model.presence.ILiteralReferencable;
import org.fortiss.variability.model.presence.PresenceCondition;
import org.fortiss.variability.model.presence.PresenceConditionTerm;

/**
 * Utility methods related to variability.
 * 
 * @author bayha
 */
public class VariabilityUtils {

	/**
	 * Regex, describing all legal characters that might be in a feature literal
	 * name.
	 */
	public final static String FEATURE_LITERAL_LEGAL_CHAR_REGEX =
			"[a-zA-Z\\d\\$&\\[\\]_\\-\\+@/\\\\,\\.=]";

	/** Regex, describing all keywords. */
	public final static String PRESENCE_CONDITIONS_KEYWORDS = "OR|AND|NOT|DEFAULT";

	/**
	 * Renames the given literal reference and updates the StringReprtesentations of
	 * the {@link PresenceCondition}s, that use it as a literal.
	 * 
	 * @param literal
	 *            The {@link ILiteralReferencable} to be renamed.
	 * @param newName
	 *            The new name.
	 */
	public static void renameLiteralReferencable(ILiteralReferencable literal, String newName) {
		String oldName = literal.getName();

		literal.setName(newName);

		Collection<PresenceConditionTerm> allReferences =
				getAllReferences(literal).stream().filter(e -> e instanceof PresenceConditionTerm)
						.map(e -> (PresenceConditionTerm)e).collect(Collectors.toList());

		for(PresenceConditionTerm r : allReferences) {
			fixStringRepresentation(r, oldName, newName);
		}
	}

	/**
	 * Replaces the literals with 'oldName' by literals with 'newName' in the String
	 * Representation.
	 * 
	 * @param newName
	 *            The new name of the literal.
	 * @param oldName
	 *            The old name of the literal.
	 * @param presCond
	 *            The {@link PresenceConditionTerm} to be fixed.
	 */
	private static void fixStringRepresentation(PresenceConditionTerm presCond, String oldName,
			String newName) {
		String oldStrRep = presCond.getStringRepresentation();

		Pattern p = Pattern.compile("(\\A|[^A-Za-z\\d])(" + oldName + ")([^A-Za-z\\d]|\\Z)");
		Matcher m = p.matcher(oldStrRep);
		if(m.find()) {
			String newStrRep = m.replaceAll("$1" + newName + "$3");
			presCond.setStringRepresentation(newStrRep);
		}

		// Recursively fix parents.
		if(presCond.eContainer() instanceof PresenceConditionTerm) {
			PresenceConditionTerm parent = (PresenceConditionTerm)presCond.eContainer();
			fixStringRepresentation(parent, oldName, newName);
		}
	}

	/**
	 * Sets the given {@link CompositionalFeatureConfiguration} to be selected.
	 * 
	 * @param fc
	 *            The {@link CompositionalFeatureConfiguration} to be selected.
	 */
	public static void selectFeatureConfiguration(CompositionalFeatureConfiguration fc) {
		EList<IFeatureConfiguration<?>> subConfigs = fc.getSubConfigurations();
		subConfigs.clear();

		EList<AbstractFeature> subFeatures = fc.getFeatureReference().getSubFeatures();
		List<IFeatureConfiguration<?>> newSubConfigs = subFeatures.stream()
				.map(f -> createFeatureConfigurationForFeature(f)).collect(toList());
		subConfigs.addAll(newSubConfigs);

		fc.setSelected(true);
	}

	/**
	 * Sets the given {@link CompositionalFeatureConfiguration} to not be selected.
	 * 
	 * @param fc
	 *            The {@link CompositionalFeatureConfiguration} to not be selected.
	 */
	public static void deselectFeatureConfiguration(CompositionalFeatureConfiguration fc) {
		// Deselect feature
		fc.getSubConfigurations().clear();
		fc.setSelected(false);
	}

	/**
	 * Checks whether the given {@link CompositionalFeatureConfiguration} is allowed
	 * to be deselected. If this is possible, deselects it.
	 * 
	 * @param fc
	 *            The {@link CompositionalFeatureConfiguration} to be checked and
	 *            deselected.
	 * 
	 * @return 'true', if fc could be deselected. 'false' otherwise.
	 */
	public static boolean
			checkAndDeselectFeatureConfiguration(CompositionalFeatureConfiguration fc) {
		AbstractCompositionalFeature feature = fc.getFeatureReference();

		if(feature != null && feature.isOptional()) {
			deselectFeatureConfiguration(fc);
			return true;
		}

		return false;
	}

	/**
	 * Checks, whether the given {@code literalName} is compatible with the presence condition
	 * syntax.
	 * 
	 * @param literalName
	 *            {@link String} with the name to be checked.
	 * @return Whether the given name is compatible with the presence condition systax.
	 */
	public static boolean isNameLegalInPresenceCondition(String literalName) {
		if(literalName.matches(PRESENCE_CONDITIONS_KEYWORDS)) {
			return false;
		}

		return literalName.matches(FEATURE_LITERAL_LEGAL_CHAR_REGEX + "*");
	}

	/**
	 * Retrieves all possible (i.e. legal) target features from the project,
	 * 'constraint' is contained in.
	 * 
	 * @param constraint
	 *            The {@link AbstractCrossFeatureConstraint} to retrieve Targets for.
	 * @return A {@link List} with all possible target features.
	 */
	public static List<AbstractFeature>
			getAllLegalConstraintTargets(AbstractCrossFeatureConstraint constraint) {
		List<AbstractFeature> features = VariabilityUtilsInternal
				.getAllElementsFromSameContainmentTree(constraint, AbstractFeature.class);
		// The feature must not have a constraint towards itself. Thus
		// it is not an option.
		// Constraints towards parents make no sense, and thus are not
		// allowed either.
		features.removeAll(getParentsWithType(constraint, AbstractFeature.class));
		return features;
	}
}
